from json import dump

from bidso.simulate import simulate_bold, simulate_ieeg, simulate_electrodes, simulate_anat
from bidso import Task

from .paths import BIDS_PATH, T1_PATH, task_ieeg, task_fmri, task_anat, elec_ct


def test_simulate_root():
    """
    TODO: this should be part of the main package
    """
    participants_tsv = BIDS_PATH / 'participants.tsv'
    with participants_tsv.open('w') as f:
        f.write('participant_id\tage\tsex\n')

        f.write(f'{task_ieeg.subject}\t30\tF\n')

    d = {
        "Name": "Dataset with simulated data",
        "BIDSVersion": "1.0.2",
        "Authors": ["Giovanni Piantoni", ],
        }

    with (BIDS_PATH / 'dataset_description.json').open('w') as f:
        dump(d, f, ensure_ascii=False, indent=' ')


def test_simulate_ieeg():
    elec = simulate_electrodes(BIDS_PATH, elec_ct)
    simulate_ieeg(BIDS_PATH, task_ieeg, elec)


def test_simulate_anat():
    simulate_anat(BIDS_PATH, task_anat, T1_PATH)


def test_simulate_fmri():
    simulate_bold(BIDS_PATH, task_fmri, T1_PATH)


def test_read_fmri():
    Task(task_fmri.get_filename(BIDS_PATH))



from bidso import (Task,
                   iEEG,
                   )
from bidso.utils import bids_mkdir
from bidso.directories import (dir_Root,
                               dir_Session,
                               )

from .paths import BIDS_PATH, task_fmri, task_ieeg


def test_directories_xxx():
    dir_Root(BIDS_PATH)

    dir_Session(bids_mkdir(BIDS_PATH, task_fmri))


def test_objects_xxx():
    Task(task_fmri.get_filename(BIDS_PATH))
    ieeg = iEEG(task_ieeg.get_filename(BIDS_PATH))
    ieeg.read_electrodes()
    assert len(ieeg.electrodes.electrodes.tsv) == 28



from json import dump
from numpy import NaN, ones, r_, stack, random

from nibabel import Nifti1Image
from nibabel import load as nload

from ..objects import Task
from ..utils import replace_extension, replace_underscore, bids_mkdir


def simulate_bold(root, task_fmri, t1):
    bids_mkdir(root, task_fmri)
    mri = nload(str(t1))

    fmri_file = task_fmri.get_filename(root)
    create_bold(mri, fmri_file, task_fmri.task)
    create_events(replace_underscore(fmri_file, 'events.tsv'))

    return Task(fmri_file)


def create_bold(mri, bold_file, taskname):
    x = mri.get_data()
    DOWNSAMPLE = 4

    af = mri.affine.copy()
    af[:3, :3] *= DOWNSAMPLE   # TODO: these should be some translation as well

    xm = stack([x[i::DOWNSAMPLE, i::DOWNSAMPLE, i::DOWNSAMPLE] for i in range(DOWNSAMPLE)], axis=-1).mean(axis=-1)
    xm[xm == 0] = NaN

    # TODO: nicer time series
    SHIFT = 3
    bold = (xm[:, :, :, None] + r_[ones(SHIFT), ones(16) * 5, ones(16), ones(16) * 5, ones(16), ones(16) * 5, ones(16 - SHIFT)])

    TR = 2.

    random.seed(100)
    bold += random.random(bold.shape) * 2
    nifti = Nifti1Image(bold.astype('float32'), af)
    nifti.header['pixdim'][4] = TR
    nifti.header.set_xyzt_units('mm', 'sec')
    nifti.to_filename(str(bold_file))

    d = {
        'RepetitionTime': TR,
        'TaskName': taskname,
        }

    json_bold = replace_extension(bold_file, '.json')
    with json_bold.open('w') as f:
        dump(d, f, ensure_ascii=False, indent=' ')


def create_events(tsv_file):

    DUR = 32
    with tsv_file.open('w') as f:
        f.write('onset\tduration\ttrial_type\n')
        is_move = True
        for i in range(6):
            trial_type = 'move' if is_move else 'rest'
            is_move = not is_move
            f.write(f'{i * DUR}\t{DUR}\t{trial_type}\n')



from .files import (file_Core,
                    file_Tsv,
                    file_Json,
                    )
from .directories import (dir_Root,
                          dir_Subject,
                          )
from .objects import (Task,
                      Electrodes,
                      iEEG,
                      )
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'VERSION')) as f:
    __version__ = f.read().strip()



from pathlib import Path

from .files import file_Tsv


class dir_Core():
    def __init__(self, dirname):
        self.dirname = Path(dirname)


class dir_Root(dir_Core):
    def __init__(self, dirname):
        super().__init__(dirname)

        self.participants = file_Tsv(self.dirname / 'participants.tsv')
        self.subjects = []
        for participant in self.participants.tsv:
            self.subjects.append(dir_Subject(self.dirname / participant['participant_id'],
                                             participant))


class dir_Subject(dir_Core):
    def __init__(self, dirname, fields):
        super().__init__(dirname)
        for k, v in fields.items():
            setattr(self, k, v)


class dir_Session(dir_Core):
    def __init__(self, dirname):
        super().__init__(dirname)



from pytest import raises

from bidso.utils import replace_extension, replace_underscore, find_extension, add_modality

from .paths import BIDS_PATH, elec_ct


def test_find_extension():
    assert find_extension(elec_ct.get_filename(BIDS_PATH)) == '.tsv'


def test_replace_extension():
    assert replace_extension('file.txt', '.bin') == 'file.bin'


def test_replace_underscore():
    assert replace_underscore('file_mod.txt', 'dat.txt') == 'file_dat.txt'


def test_modality():
    assert BIDS_PATH == add_modality(BIDS_PATH, None)
    assert BIDS_PATH / 'xxx' == add_modality(BIDS_PATH, 'xxx')

    with raises(ValueError):
        add_modality(BIDS_PATH, 'events')



from bidso.files import file_Tsv
from bidso import Electrodes
from .paths import elec_ct, BIDS_PATH


def test_file_Tsv_get():

    tsv = file_Tsv(elec_ct.get_filename(BIDS_PATH))

    assert tsv.get(lambda x: x['name'] == 'grid01')[0]['name'] == 'grid01'
    assert tsv.get(None, lambda x: x['name'])[0] == 'grid01'


def test_Electrodes_get_xyz():
    elec = Electrodes(elec_ct.get_filename(BIDS_PATH))

    assert len(elec.get_xyz()) == 28
    assert len(elec.get_xyz(['grid01', ])) == 1



from pathlib import Path
from bidso import file_Core


TEST_PATH = Path(__file__).resolve().parent
DATA_PATH = TEST_PATH / 'data'
BIDS_PATH = DATA_PATH / 'bids'
BIDS_PATH.mkdir(parents=True, exist_ok=True)
DERIVATIVES_PATH = DATA_PATH / 'derivatives'
FREESURFER_PATH = DERIVATIVES_PATH / 'freesurfer'

subject = 'bert'
task_ieeg = file_Core(
    subject=subject,
    session='day02',
    modality='ieeg',
    task='block',
    run='1',
    acquisition='clinical',
    extension='.eeg',
    )
task_fmri = file_Core(
    subject=subject,
    session='day01',
    modality='bold',
    task='block',
    run='1',
    extension='.nii.gz',
    )
task_anat = file_Core(
    subject=subject,
    session='day01',
    modality='T1w',
    extension='.nii.gz',
    )
elec_ct = file_Core(
    subject=subject,
    session='day02',
    modality='electrodes',
    acquisition='ct',
    extension='.tsv',
    )


T1_PATH = FREESURFER_PATH / 'sub-bert/mri/T1.mgz'



from json import dump
from textwrap import dedent
from datetime import datetime
from shutil import copyfile
from pathlib import Path
from numpy import ones, memmap, r_
from numpy import random

import bidso

from ..objects import Electrodes, iEEG
from ..utils import replace_underscore, replace_extension, bids_mkdir
from .fmri import create_events


DATA_PATH = Path(__file__).resolve().parent / 'data'

S_FREQ = 256
DURATION = 192
AMPLITUDE = 1000
EFFECT_SIZE = 2
block_dur = 32
EXTRA_CHANS = ['EOG1', 'EOG2', 'ECG', 'EMG', 'other']

BV_ORIENTATION = 'MULTIPLEXED'  # 'F'
BV_DATATYPE = 'IEEE_FLOAT_32'  # float32
RESOLUTION = 1

fake_time = datetime(2017, 5, 16, 20, 45, 6)


def simulate_ieeg(root, ieeg_task, elec):
    bids_mkdir(root, ieeg_task)

    chan_names = [x['name'] for x in elec.electrodes.tsv]
    ieeg_file = ieeg_task.get_filename(root)

    create_ieeg_data(ieeg_file, chan_names)
    create_ieeg_info(replace_extension(ieeg_file, '.json'))
    create_channels(replace_underscore(ieeg_file, 'channels.tsv'), elec)
    create_events(replace_underscore(ieeg_file, 'events.tsv'))

    return iEEG(ieeg_file)


def simulate_electrodes(root, elec_obj, electrodes_file=None):
    bids_mkdir(root, elec_obj)

    if electrodes_file is None:
        electrodes_file = DATA_PATH / 'electrodes.tsv'
    output_file = elec_obj.get_filename(root)
    copyfile(electrodes_file, output_file)

    coordsystem_file = replace_underscore(output_file, 'coordsystem.json')
    COORDSYSTEM = {
        "iEEGCoordinateSystem": 'T1w',
        "iEEGCoordinateUnits": 'mm',
        "iEEGCoordinateProcessingDescription": "none",
        "IntendedFor": "/sub-bert/ses-day01/anat/sub-bert_ses-day01_acq-wholebrain_T1w.nii.gz",
        "AssociatedImageCoordinateSystem": "T1w",
        "AssociatedImageCoordinateUnits": "mm",
        }

    with coordsystem_file.open('w') as f:
        dump(COORDSYSTEM, f, indent=' ')

    return Electrodes(output_file)


def create_ieeg_data(output_file, elecs):

    _write_vhdr(output_file.with_suffix('.vhdr'), elecs + EXTRA_CHANS)
    _write_vmrk(output_file.with_suffix('.vmrk'))

    n_chan = len(elecs) + len(EXTRA_CHANS)

    random.seed(100)
    t = r_[ones(block_dur * S_FREQ) * EFFECT_SIZE, ones(block_dur * S_FREQ), ones(block_dur * S_FREQ) * EFFECT_SIZE, ones(block_dur * S_FREQ), ones(block_dur * S_FREQ) * EFFECT_SIZE, ones(block_dur * S_FREQ)]
    data = random.random((n_chan, S_FREQ * DURATION)) * t[None, :] * AMPLITUDE

    dtype = 'float32'
    memshape = (n_chan, S_FREQ * DURATION)
    mem = memmap(str(output_file), dtype, mode='w+', shape=memshape, order='F')
    mem[:, :] = data
    mem.flush()


def create_channels(output_file, elec):
    with output_file.open('w') as f:
        f.write('name\ttype\tunits\tsampling_frequency\tlow_cutoff\thigh_cutoff\tnotch\treference\tstatus\n')
        for one_elec in elec.electrodes.tsv:
            f.write(f'{one_elec["name"]}\tECOG\tµV\t{S_FREQ}\tn/a\tn/a\tn/a\tn/a\tgood\n')

        for chan_name in EXTRA_CHANS:
            f.write(f'{chan_name}\tother\tµV\t{S_FREQ}\tn/a\tn/a\tn/a\tn/a\tgood\n')


def create_ieeg_info(output_file):
    """Use only required fields
    """
    dataset_info = {
        "TaskName": "block",
        "Manufacturer": "simulated",
        "PowerLineFrequency": 50,
    }

    with output_file.open('w') as f:
        dump(dataset_info, f, indent=' ')


def _write_vhdr(filename, chans):
    vhdr_txt = f"""\
    Brain Vision Data Exchange Header File Version 1.0
    ; Data created by the BIDSO {bidso.__version__} on {fake_time}

    [Common Infos]
    Codepage=UTF-8
    DataFile={filename.stem}.eeg
    MarkerFile={filename.stem}.vmrk
    DataFormat=BINARY
    ; Data orientation: MULTIPLEXED=ch1,pt1, ch2,pt1 ...
    DataOrientation={BV_ORIENTATION}
    NumberOfChannels={len(chans)}
    ; Sampling interval in microseconds
    SamplingInterval={1e6 / S_FREQ:f}

    [Binary Infos]
    BinaryFormat={BV_DATATYPE}

    [Channel Infos]
    ; Each entry: Ch<Channel number>=<Name>,<Reference channel name>,
    ; <Resolution in "Unit">,<Unit>, Future extensions..
    ; Fields are delimited by commas, some fields might be omitted (empty).
    """
    # found a way to write \1
    vhdr_txt += r'; Commas in channel names are coded as "\1".'
    vhdr_txt += '\n'

    output = []
    for i, chan in enumerate(chans):
        output.append(f'Ch{i + 1:d}={chan},,{RESOLUTION},µV')

    with filename.open('w') as f:
        f.write(dedent(vhdr_txt) + '\n'.join(output))


def _write_vmrk(filename):

    vmrk_txt = f"""\
    Brain Vision Data Exchange Marker File, Version 1.0

    [Common Infos]
    Codepage=UTF-8
    DataFile={filename.name}

    [Marker Infos]
    ; Each entry: Mk<Marker number>=<Type>,<Description>,<Position in data points>,
    ; <Size in data points>, <Channel number (0 = marker is related to all channels)>
    ; Fields are delimited by commas, some fields might be omitted (empty).
    """
    vmrk_txt = dedent(vmrk_txt)
    # found a way to write \1
    vmrk_txt += r'; Commas in type or description are coded as "\1".'
    vmrk_txt += f'\nMk1=New Segment,,1,1,0,{fake_time:%Y%m%d%H%M%S%f}\n'

    with filename.open('w') as f:
        f.write(vmrk_txt)



from pytest import raises
from bidso import file_Core
from .paths import BIDS_PATH, elec_ct


def test_create_file_Core():
    core = file_Core(subject='a', session='b')
    assert core.subject == 'a'
    assert core.run is None

    with raises(AttributeError):
        core = file_Core(weird='a')


def test_get_filename():
    core = file_Core(
        subject='bert',
        session='day02',
        task='block',
        run='00',
        acquisition='ct',
        modality='ieeg')
    expected = 'sub-bert_ses-day02_task-block_run-00_acq-ct_ieeg'
    assert core.get_filename() == expected
    assert core.get_filename(BIDS_PATH).name == expected
    assert str(core.get_filename(BIDS_PATH).relative_to(BIDS_PATH).parent) == 'sub-bert/ses-day02/ieeg'

    assert core.get_filename(BIDS_PATH).parent.name == 'ieeg'
    assert core.get_filename(BIDS_PATH, 'xxx').parent.name == 'xxx'


def test_get_filename_modality():
    core = file_Core(
        subject='bert',
        session='day02',
        task='block',
        run='00',
        acquisition='ct',
        modality='ieeg')

    assert core.get_filename(BIDS_PATH).parent.name == 'ieeg'
    assert core.get_filename(BIDS_PATH, 'xxx').parent.name == 'xxx'


def test_elec_filename():
    assert elec_ct.get_filename() == 'sub-bert_ses-day02_acq-ct_electrodes.tsv'



from warnings import warn
from logging import getLogger
from re import sub

lg = getLogger(__name__)


def find_root(filename, target='bids'):
    """Find base directory (root) for a filename.

    Parameters
    ----------
    filename : instance of Path
        search the root for this file
    target: str
        'bids' (the directory containing 'participants.tsv'), 'subject' (the
        directory starting with 'sub-'), 'session' (the directory starting with
        'ses-')

    Returns
    -------
    Path
        path of the target directory
    """
    lg.debug(f'Searching root in {filename}')
    if target == 'bids' and (filename / 'dataset_description.json').exists():
        return filename
    elif filename.is_dir():
        pattern = target[:3] + '-'
        if filename.stem.startswith(pattern):
            return filename

    return find_root(filename.parent, target)


def find_in_bids(filename, pattern=None, generator=False, upwards=False,
                 wildcard=True, **kwargs):
    """Find nearest file matching some criteria.

    Parameters
    ----------
    filename : instance of Path
        search the root for this file
    pattern : str
        glob string for search criteria of the filename of interest (remember
        to include '*'). The pattern is passed directly to rglob.
    wildcard : bool
        use wildcards for unspecified fields or not (if True, add "_*_" between
        fields)
    upwards : bool
        where to keep on searching upwards
    kwargs : dict


    Returns
    -------
    Path
        filename matching the pattern
    """
    if upwards and generator:
        raise ValueError('You cannot search upwards and have a generator')

    if pattern is None:
        pattern = _generate_pattern(wildcard, kwargs)

    lg.debug(f'Searching {pattern} in {filename}')

    if upwards and filename == find_root(filename):
        raise FileNotFoundError(f'Could not find file matchting {pattern} in {filename}')

    if generator:
        return filename.rglob(pattern)

    matches = list(filename.rglob(pattern))
    if len(matches) == 1:
        return matches[0]

    elif len(matches) == 0:
        if upwards:
            return find_in_bids(filename.parent, pattern=pattern, upwards=upwards)
        else:
            raise FileNotFoundError(f'Could not find file matchting {pattern} in {filename}')

    else:
        matches_str = '"\n\t"'.join(str(x) for x in matches)
        raise FileNotFoundError(f'Multiple files matching "{pattern}":\n\t"{matches_str}"')


def _generate_pattern(wildcard, kwargs):

    if not wildcard and 'subject' not in kwargs:
        raise ValueError('You need to specify "subject" if you do not use any wildcard')

    unknown = set(kwargs) - set(['subject', 'session', 'task', 'run', 'acquisition', 'modality', 'extension'])
    if len(unknown) > 0:
        warn('Unrecognized type ' + ', '.join([f'"{x}"' for x in unknown]))

    if 'subject' in kwargs:
        pattern = 'sub-' + kwargs['subject']
    else:
        pattern = '*'

    if 'session' in kwargs:
        pattern += '_ses-' + kwargs['session'] + '_'
    elif wildcard:
        pattern += '_*'

    if 'task' in kwargs:
        pattern += '_task-' + kwargs['task'] + '_'
    elif wildcard:
        pattern += '_*'

    if 'run' in kwargs:
        pattern += '_run-' + kwargs['run'] + '_'
    elif wildcard:
        pattern += '_*'

    if 'acquisition' in kwargs:
        pattern += '_acq-' + kwargs['acquisition'] + '_'
    elif wildcard:
        pattern += '_*'

    if 'modality' in kwargs:
        pattern += '_' + kwargs['modality']
    elif wildcard:
        pattern += '_*'

    if 'extension' in kwargs:
        pattern += kwargs['extension']
    else:
        pattern += '.*'

    pattern = sub('(_\*)+', '_*', pattern)
    pattern = sub('(\*_)+', '*_', pattern)
    pattern = pattern.replace('__', '_')

    return pattern



from .files import file_Core, file_Json, file_Tsv
from .utils import replace_extension, replace_underscore
from .find import find_in_bids


class Electrodes(file_Core):
    def __init__(self, filename):
        super().__init__(filename)
        self.electrodes = file_Tsv(filename)
        self.coordframe = file_Json(replace_underscore(self.filename, 'coordsystem.json'))

    def get_xyz(self, list_of_names=None):
        """Get xyz coordinates for these electrodes

        Parameters
        ----------
        list_of_names : list of str
            list of electrode names to use

        Returns
        -------
        list of tuples of 3 floats (x, y, z)
            list of xyz coordinates for all the electrodes

        TODO
        ----
        coordinate system of electrodes
        """
        if list_of_names is not None:
            filter_lambda = lambda x: x['name'] in list_of_names
        else:
            filter_lambda = None

        return self.electrodes.get(filter_lambda=filter_lambda,
                                   map_lambda=lambda e: (float(e['x']),
                                                         float(e['y']),
                                                         float(e['z'])))


class Task(file_Core):
    def __init__(self, filename):
        super().__init__(filename)
        self.events = file_Tsv(replace_underscore(self.filename, 'events.tsv'))
        self.json = file_Json(replace_extension(self.filename, '.json'))


class iEEG(Task):
    electrodes = None
    channels = None

    def __init__(self, filename):
        super().__init__(filename)
        self.channels = file_Tsv(replace_underscore(self.filename, 'channels.tsv'))

    def read_electrodes(self, electrodes='*'):
        self.electrodes = Electrodes(find_in_bids(self.filename, upwards=True, acquisition=electrodes, modality='electrodes', extension='.tsv'))



from logging import getLogger
from pathlib import Path
from re import search

lg = getLogger(__name__)


def read_tsv(filename):
    filename = Path(filename)
    with filename.open() as f:
        hdr = f.readline()
        tsv = []
        for l in f:
            d = {k.strip(): v.strip() for k, v in zip(hdr.split('\t'), l.split('\t'))}
            tsv.append(d)
    return tsv


def find_extension(filename):
    if isinstance(filename, Path):
        filename = filename.name

    return '.' + '.'.join(filename.split('.')[1:])


def add_extension(filename, suffix):
    if isinstance(filename, str):
        return filename + suffix
    else:
        return filename.parent / (filename.name + suffix)


def replace_extension(filename, suffix):
    return add_extension(remove_extension(filename), suffix)


def remove_extension(filename):
    if isinstance(filename, str):
        return filename.split('.')[0]
    else:
        return filename.parent / filename.name.split('.')[0]


def add_underscore(filename, suffix):
    if isinstance(filename, str):
        return filename + '_' + suffix
    else:
        return filename.parent / (filename.name + '_' + suffix)


def replace_underscore(filename, suffix):
    return add_underscore(remove_underscore(filename), suffix)


def remove_underscore(filename):
    if isinstance(filename, str):
        return '_'.join(filename.split('_')[:-1])
    else:
        return filename.parent / ('_'.join(filename.name.split('_')[:-1]))


def bids_mkdir(base_path, file_bids):

    output_path = base_path / ('sub-' + file_bids.subject)
    output_path.mkdir(exist_ok=True)
    if file_bids.session is not None:  # hasattr for pybids, isnone for bidso
        output_path = output_path / ('ses-' + file_bids.session)
        output_path.mkdir(exist_ok=True)

    output_path = add_modality(output_path, file_bids.modality)
    output_path.mkdir(exist_ok=True)

    return output_path


def add_modality(output_path, modality):
    """Modality can be appended to the file name (such as 'bold') or use in the
    folder (such as "func"). You should always use the specific modality ('bold').
    This function converts it to the folder name.
    """
    if modality is None:
        return output_path
    else:
        if modality == 'T1w':
            modality = 'anat'

        if modality == 'bold':
            modality = 'func'

        if modality in ('electrodes', 'coordsystem', 'channels'):
            modality = 'ieeg'

        if modality == 'events':
            raise ValueError('modality "events" is ambiguous (can be in folder "ieeg" or "func"). Assuming "ieeg"')

        return output_path / modality


def _match(filename, pattern):
    m = search(pattern, filename.stem)
    if m is None:
        return m
    else:
        return m.group(1)



from json import load as json_load
from pathlib import Path

from .utils import read_tsv, _match, find_extension, add_modality, remove_extension


class file_Core():
    filename = None
    subject = None
    session = None
    modality = None
    run = None
    acquisition = None
    task = None
    extension = None

    def __init__(self, filename=None, **kwargs):
        if filename is not None:
            self.filename = Path(filename)
            self.subject = _match(self.filename, 'sub-([a-zA-Z0-9\-]+)_')
            self.session = _match(self.filename, '_ses-([a-zA-Z0-9\-]+)_')
            self.modality = remove_extension(self.filename.name).split('_')[-1]
            self.run = _match(self.filename, '_run-([a-zA-Z0-9\-]+)_')
            self.acquisition = _match(self.filename, '_acq-([a-zA-Z0-9\-]+)_')
            self.task = _match(self.filename, '_task-([a-zA-Z0-9\-]+)_')
            self.extension = find_extension(self.filename)

        else:
            for k, v in kwargs.items():
                if hasattr(self, k):
                    setattr(self, k, v)
                else:
                    raise AttributeError(f'"{k}" is not an attribute of file_Core')

    def get_filename(self, base_dir=None, modality=None):
        """Construct filename based on the attributes.

        Parameters
        ----------
        base_dir : Path
            path of the root directory. If specified, the return value is a Path,
            with base_dir / sub-XXX / (ses-XXX /) modality / filename
            otherwise the return value is a string.
        modality : str
            overwrite value for modality (i.e. the directory inside subject/session).
            This is necessary because sometimes the modality attribute is ambiguous.

        Returns
        -------
        str or Path
            str of the filename if base_dir is not specified, otherwise the full
            Path
        """
        filename = 'sub-' + self.subject
        if self.session is not None:
            filename += '_ses-' + self.session
        if self.task is not None:
            filename += '_task-' + self.task
        if self.run is not None:
            filename += '_run-' + self.run
        if self.acquisition is not None:
            filename += '_acq-' + self.acquisition
        if self.modality is not None:
            filename += '_' + self.modality
        if self.extension is not None:
            filename += self.extension

        if base_dir is None:
            return filename

        else:
            dir_name = base_dir / ('sub-' + self.subject)
            if self.session is not None:
                dir_name /= 'ses-' + self.session

            if modality is not None:
                dir_name /= modality
            else:
                dir_name = add_modality(dir_name, self.modality)

            return dir_name / filename


class file_Tsv(file_Core):
    def __init__(self, filename):
        super().__init__(filename)
        self.tsv = read_tsv(self.filename)

    def get(self, filter_lambda=None, map_lambda=None):
        """Select elements of the TSV, using python filter and map.

        Parameters
        ----------
        filter_lambda : function
            function to filter the tsv rows (the function needs to return True/False)
        map_lambda : function
            function to select the tsv columns

        Returns
        -------
        list
            list (not a generator, because that's the most common case)

        Examples
        --------
        To select all the channels in one list, called "good_labels"::

            >>> file_Tsv.get(lambda x: x['name'] in good_labels)

        To select all the names of the channels:

            >>> file_Tsv.get(map_filter=lambda x: x['name'])

        """
        if filter_lambda is None:
            filter_lambda = lambda x: True
        if map_lambda is None:
            map_lambda = lambda x: x
        return list(map(map_lambda, filter(filter_lambda, self.tsv)))


class file_Json(file_Core):
    def __init__(self, filename):
        super().__init__(filename)
        with self.filename.open() as f:
            self.json = json_load(f)



from pytest import raises, warns
from bidso.find import find_in_bids, find_root, _generate_pattern

from .paths import BIDS_PATH, task_ieeg

filename = task_ieeg.get_filename(BIDS_PATH)


def test_find_root():
    assert find_root(filename).name == 'bids'
    assert find_root(filename, target='subject').name == 'sub-bert'
    assert find_root(filename, target='session').name == 'ses-day02'


def test_find_in_bids_01():

    found = find_in_bids(filename, subject='bert', session='day01', run='1',
                         extension='.nii.gz', upwards=True)
    assert found.name == 'sub-bert_ses-day01_task-block_run-1_bold.nii.gz'

    with warns(UserWarning):
        find_in_bids(filename, subject='bert', useless='xxx', task='block',
                     modality='channels', upwards=True)


def test_find_in_bids_02():

    with raises(FileNotFoundError):
        find_in_bids(filename, upwards=True, subject='xxx')


def test_find_in_bids_03():

    with raises(FileNotFoundError):
        find_in_bids(filename, upwards=True, subject='bert')


def test_find_in_bids_04():
    assert sum(1 for x in find_in_bids(BIDS_PATH, generator=True, subject='bert')) == 12


def test_find_in_bids_05():
    with raises(FileNotFoundError):
        find_in_bids(BIDS_PATH, subject='xxx')

    with raises(StopIteration):
        next(find_in_bids(BIDS_PATH, subject='xxx', generator=True))


def test_find_in_bids_06():
    with raises(ValueError):
        find_in_bids(BIDS_PATH, upwards=True, generator=True)


def test_generate_pattern():
    assert _generate_pattern(True, dict(subject='test')) == 'sub-test_*.*'
    assert _generate_pattern(False, dict(subject='test')) == 'sub-test.*'

    assert _generate_pattern(True, dict(subject='test', session='sess')) == 'sub-test_ses-sess_*.*'
    assert _generate_pattern(True, dict(subject='test', modality='mod')) == 'sub-test_*_mod.*'
    assert _generate_pattern(True, dict(session='sess', extension='.nii.gz')) == '*_ses-sess_*.nii.gz'
    assert _generate_pattern(True, dict(modality='mod', extension='.nii.gz')) == '*_mod.nii.gz'


def test_wildcard_subject():
    with raises(ValueError):
        _generate_pattern(False, dict())
